"""
pyodide-mkdocs-theme
Copyleft GNU GPLv3 🄯 2024 Frédéric Zinelli

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.
If not, see <https://www.gnu.org/licenses/>.
"""

# pylint: disable=unused-argument


import re
import hashlib
from typing import Any, ClassVar, Dict, List, Literal, Optional, Tuple, Union
from dataclasses import dataclass
from pathlib import Path
from math import inf

from mkdocs.exceptions import BuildError

from pyodide_mkdocs_theme.pyodide_macros.messages import Tip
from pyodide_mkdocs_theme.pyodide_macros.paths_utils import get_ide_button_png_path

from .. import html_builder as Html
from ..pyodide_logger import logger
from ..tools_and_constants import HtmlClass, IdeConstants, ScriptKind
from ..parsing import items_comma_joiner
from ..plugin.maestro_IDE import MaestroIDE

from .ide_files_data import IdeFilesExtractor




#---------------------------------------------------------------------------------





@dataclass
class IdeManager:
    """
    Base class managing the information for the underlying environment.
    To be extended by a concrete implementation, providing the actual logistic to
    build the html hierarchy (see self.make_element).

    This base class holds all the possible arguments fo any concrete class.
    """


    # Defined on instantiation:
    #--------------------------

    env: MaestroIDE
    """ The MaestroEnv singleton """

    py_name: str
    """ Base name for the files to use (first argument passed to the macros)
        Partial path from the directory holding the sujet.md file, to the one holding all the
        other required files, ending with the common prefix for the exercice.
        Ex:   "exo" to extract:   "exo.py", "exo_corr.py", "exo_test.py", ...
                "sub_exA/exo" for:  "sub_exA/exo.py", "sub_exA/exo_corr.py", ...
    """

    id: Optional[int]
    """ Used to disambiguate the ids of two IDEs, if the same file is used several times
        in the document.
    """

    excluded: str
    """ String of spaces or coma separated python functions or modules/packages that are forbidden
        at runtime. By default, nothing is forbidden.
            - Every string section that matches a builtin callable forbid that function by
              replacing it with another function which will raise an error if called.
            - Every string section prefixed with a fot forbids a method call. Here a simple
              string containment check is done opn the user's code, to check it does not
              contain the desired method name with the dot before it.
            - Any other string section is considered as a module name and doing an import (in
              any way/syntax) involving that name will raise an error.

        Note that the restrictions are rather strict, and may have unexpected side effects, such
        as, forbidding `exec` will also forbid to import numpy, because the package relies on exec
        for some operations at import time.
        To circumvent such a kind of problems, use the white_list argument.
    """

    white_list: str
    """ String of spaces or coma separated python modules/packages names the have to be
        preloaded before the code restrictions are enforced on the user's side.
    """

    rec_limit: int
    """ If used, the recursion limit of the pyodide runtime will be updated before the user's
        code or the tests are run.
        Note that this also forbids the use of the `sys.setrecurionlimit` at runtime.
    """

    mode: Union[Literal[""],Literal["_v"]] = ""
    """ The terminal will be below (mode="") or on the right (mode="_v") of the editor.
        (what an awful interface, yeah... x) )
    """

    max_attempts: Optional[Union[int, Literal["+"]]] = None
    """ Maximum number of attempts before the solution admonition will become available.
        If None, use the global default value.
    """

    max_size: Optional[int] = None
    """
    Max height of the editor (in number of lines)
    """

    auto_log_assert: Optional[bool] = None
    """ If True, failing assertions without feedback during the private tests will be
        augmented automatically with the code of the assertion itself. If None, use
        the global `show_assertion_code_on_failed_test` plugin value and defined in
        `CONFIG.showAssertionCodeOnFailedTest` in the JS runtime.
    """

    term_height: Optional[int] = None
    """
    Number of lines to define the height of the terminal (unless it's vertical)
    """

    prefill_term: Optional[str] = None
    """
    Command to prefill the terminal on startup.
    """

    extra_kw: Optional[Dict[str,Any]] = None
    """
    Any kw left in the original call.
    Should be always be None when reaching IdeManager.__post_init__. This allows subclasses
    to handle the extra (legacy) keywords on their side.
    """


    # defined during post_init
    #-------------------------


    files_data: IdeFilesExtractor = None

    editor_name: str = ''
    """ tail part of most ids, in the shape of 'editor_{32 bits hexadecimal}' """

    max_attempts_symbol: str = ''
    """ Actual string representation to use when creating the counter under the IDE """

    indentation: str = ''
    """ Indentation on the left of the macro call, as str """


    #-------------------------


    ID_PREFIX: ClassVar[str] = None
    """ Must be overridden in the child class """

    NEED_INDENTS: ClassVar[bool] = False
    """
    Specify the macro had adding multiline content (so it _will_ consume one indentation data).
    """

    NEEDED_KINDS: ClassVar[Tuple[ScriptKind]] = ScriptKind.pyodide,
    """
    Register the kind of js scripts that must be added to the page, for the current object.
    """

    KW_TO_TRANSFER: ClassVar[Tuple[ Union[str, Tuple[str,str]]] ]  = ()
    """
    Configuration of the keywords that should be extracted if given in the constructor.
    This is an Iterable of  (argument, property) pairs.
    If an element is a string, it will be used as (value, value.lower())
    """

    DEFAULTS_EXTRACTION: ClassVar[Tuple[Tuple[str,str]]] = ()
    """
    Configuration of the properties that should extract a global default value, if it's still
    set to None after the extra arguments have been handled.
    This is an Iterable of  (argument, property) pairs.
    """

    JS_EXPORTED_GENERICS: ClassVar[set] = set('''
        py_name
        excluded
        excluded_methods
        rec_limit
        white_list
    '''.split() + [
        prop for prop in IdeFilesExtractor.SECTION_TO_PROP.values() if prop !="corr_content"
    ])
    """
    Values that are always exported to JS, whatever the tool.
    """

    JS_EXPORTED_VALUES: ClassVar[set] = None
    """
    If defined, gives the names of all the properties that should be exported to JS.
    HAS TO BE OVERRIDDEN by the subclasses.
    """


    #-------------------------


    @property               # pylint: disable-next=all
    def has_corr(self):     return self.files_data.has_corr
    @property               # pylint: disable-next=all
    def has_secrets(self):  return self.files_data.has_secrets
    @property               # pylint: disable-next=all
    def has_rem(self):      return self.files_data.has_rem
    @property               # pylint: disable-next=all
    def has_vis_rem(self):  return self.files_data.has_vis_rem

    @property               # pylint: disable-next=all
    def has_any_corr_rems(self):
        return self.has_corr or self.has_rem or self.has_vis_rem



    def __post_init__(self):

        self.handle_extra_args()

        self.env.set_current_page_insertion_needs(*self.NEEDED_KINDS)

        self.files_data = IdeFilesExtractor(self.env, self.py_name)

        # Extract max number of attempts from file or macro argument, clean up the file if needed,
        # then pick the correct number of attempts and set it in the global structure.
        # Also defines self.ide_content.
        max_attempts = self._define_max_attempts_symbols_and_value()

        self._validate_files_config(max_attempts)

        self.has_check_btn   = self.has_secrets # or self.has_any_corr_rem and cnt_n_span != inf

        if self.rec_limit < -1:         # standardization
            self.rec_limit = -1

        if -1 < self.rec_limit < IdeConstants.min_recursion_limit:
            raise BuildError(
                f"The recursion limit for {self} is set too low and may causes runtime troubles. "
                f"Please set it to at least { IdeConstants.min_recursion_limit }."
            )

        # Extract python content and compute editor name:
        if self.ID_PREFIX is None:
            raise NotImplementedError("Subclasses should override the ID_PREFIX class property.")


        self.editor_name = self.generate_id()


        # Compute all code exclusions and white list of imports:
        white_list = self._compute_exclusions_and_white_list("white_list")
        excluded = self._compute_exclusions_and_white_list("excluded")
        excluded_methods = [ meth for meth in excluded if     meth.startswith('.') ]
        excluded =         [ meth for meth in excluded if not meth.startswith('.') ]


        # Search the indentation level for the current IDE:
        if self.NEED_INDENTS:
            self.indentation = self.env.get_macro_indent()

        # Guess an explicative enough py_name (when downloading the IDE content):
        root_name = Path(self.env.page.url).stem
        py_path = Path(self.py_name).stem
        py_name = f"{root_name}-{py_path}".strip('-') or 'unknown'

        # Prepare the dump of configuration to JS:
        to_register: List[Tuple[str,Any]] = [
            #-----------------------------------------------------
            # Generic properties (always exported, but may be useless depending on the actual class)
            ('py_name',             py_name+'.py'),
            ("excluded",            excluded),
            ("excluded_methods",    excluded_methods),
            ("rec_limit",           self.rec_limit),
            ("white_list",          white_list),
            *self.files_data.get_sections_data(with_corr=False),

            #-----------------------------------------------------
            # Specific to some concrete class

            # IDEs:
            ('attempts_left',       max_attempts),
            ("auto_log_assert",     self.auto_log_assert),
            ('corr_rems_mask',      self.files_data.corr_rems_bit_mask),
            ("has_check_btn",       self.has_check_btn),

            # Terminals:
            ("prefill_term",        self.prefill_term),
        ]

        actually_registered = [
            rule for rule in to_register
                 if rule[0] in self.JS_EXPORTED_GENERICS
                 or rule[0] in self.JS_EXPORTED_VALUES
        ]
        for field,value in actually_registered:
            self.env.set_current_page_js_data(self.editor_name, field, value)



    #-----------------------------------------------------------------------------


    def __str__(self):
        return f"{self.__class__.__name__}('{self.py_name}', ...), in file {self.env.location()}"



    def handle_extra_args(self):
        """
        Fill the extra arguments provided through other keyword arguments, handling only those
        actually required for the child class. If some are remaining, after this in self.extra_kw,
        an error will be raised.
        Also extract default values for properties that are still set to None after handling the
        keyword arguments.
        """
        to_transfer = [
            data if isinstance(data,tuple) else (data,data.lower()) for data in self.KW_TO_TRANSFER
        ]
        for kw, prop in to_transfer:
            if kw in self.extra_kw:
                value = self.extra_kw.pop(kw)
                setattr(self, prop, value)

        if self.extra_kw:
            raise BuildError(
                f"Fund forbidden arguments for {self}:\n" + "".join(
                    f"    {k} = {v!r}\n" for k,v in self.extra_kw.items()
                )
            )

        self.get_default_if_none(self.DEFAULTS_EXTRACTION)



    def get_default_if_none(self, props_to_options: Tuple[str,str]):
        """ Extract the default values if the current property is set to None. """
        for prop, option_prop in props_to_options:
            if getattr(self, prop) is None:
                default_value = getattr(self.env, option_prop)
                setattr(self, prop, default_value)



    def _define_max_attempts_symbols_and_value(self):
        """
        Any MAX value defined in the file takes precedence, because it's not possible to know
        if the value coming from the macro is the default one or not.
        """
        max_ide = str(self.max_attempts)        # from macro call

        # If something about MAX in the file, it has precedence (if exists -> legacy...)
        max_from_file = self.files_data.file_max_attempts
        if max_from_file != "":
            max_ide = max_from_file

        is_inf = max_ide in ("+", "1000")     # 1000: legacy...

        # If ever there are neither correction nor remark, or if no tests, use also inf:
        is_inf = is_inf or not self.has_any_corr_rems or not self.has_secrets

        self.max_attempts_symbol = IdeConstants.infinity_symbol if is_inf else str(max_ide)

        max_attempts = inf if is_inf else int(max_ide)
        return max_attempts



    def _validate_files_config(self, max_attempts: Union[int,float] ):
        raise NotImplementedError()



    def _validation_outcome(self, msg:Optional[str], config_opt:Optional[str]=None):
        if not msg:
            return

        msg = f"Invalid configuration with: {self}\n    {msg}"
        if config_opt:
            msg += (
                "\n    You can deactivate this check by setting plugins.pyodide_macros."
                f"build.{config_opt} to false in `mkdocs.yml`."
            )

        if self.env._dev_mode:       # pylint: disable=protected-access
            logger.error("DEV_MODE (expected x5) - " + msg)
        else:
            raise BuildError(msg)



    def _compute_exclusions_and_white_list(self, prop:str):
        """
        Convert a string argument (exclusions or white list) tot he equivalent list of data.
        """
        rule = (getattr(self, prop) or "").strip(' ;,')       # (never allow None)
        lst = re.split(r'[ ;,]+', rule) if rule else []
        return lst



    def generate_id(self):
        """
        Generate an id number for the current element, in the form:

            PREFIX_{32 bits hash value}

        This id must be:
            - Unique to every IDE used throughout the whole website.
            - Stable, so that it can be used to identify what IDE goes with what file or what
              localStorage data.

        Current strategy:
            - If the file exists, hash its path.
            - If there is no file, use the current global IDE_counter and hash its value as string.
            - The "mode" of the IDE is appended to the string before hashing.
            - Any ID value (macro argument) is also appended to the string before hashing.

        Uniqueness of the resulting hash is verified and a BuildError is raised if two identical
        hashes are encountered.

        NOTE: uniqueness most be guaranteed for IDEs (LocalStorage). It's less critical for other
        elements, but they still need to stay unique across a page, at least (especially when
        feedback is involved, like with terminals. Note: maybe not anymore... :thinking: )
        """
        path = str(self.env.generic_count)
        if self.id is not None:
            path += str(self.id)            # kept in case unlucky collision... (yeah, proba... XD )
        return self.id_to_hash(path)



    def id_to_hash(self, clear:str):
        """ Hash the hash to add as html id tail, prefix it, and check the uniqueness of the hash
            across the whole website.
        """

        hashed  = hashlib.sha1(clear.encode("utf-8")).hexdigest()
        html_id = f"{ self.ID_PREFIX }{ hashed }"

        if not self.env.register_if_unique(html_id):
            raise BuildError(
                "The same html id got generated twice.\nIf you are trying to use the same set "
                "of python files for different IDE-related elements, use their ID argument to "
                "disambiguate their id.\n"
               f"  Problematic call:  { self }\n"
               f"  Possible solution: add the ID:int keyword argument or change its current value"
            )
        return html_id



    #-----------------------------------------------------------------------------



    def make_element(self) -> str:
        """
        Create an IDE (Editor+Terminal) within an Mkdocs document. {py_name}.py is loaded
        on the editor if present.
        NOTES:
            - Two modes are available : vertical or horizontal. Buttons are added through
                functional calls.
            - The last span hides the code content of the IDE when loaded.
        """
        raise NotImplementedError("Subclasses should implement the make_element method.")



    def create_button(
        self,
        button_name:    str,
        *,
        btn_kind:       str   = None,
        margin_left:    float = 0.2,
        margin_right:   float = 0.2,
        extra_content:  str   = "",
        **kwargs
    ) -> str:
        """
        Build one button, given its name.

        @btn_kind:      The name of the JS function to bind the button click event to.
                        If none given, use the lowercase version of @button_name.
        @margin_...:    CSS formatting as floats (default: 0.2em on each side).
        @extra_content: Allow to inject some additional html inside the button tag.
        @**kwargs:      All the remaining kwargs are attributes added to the button tag.
        """
        if btn_kind is None:
            btn_kind = button_name.lower()

        style = f"margin-left:{margin_left}em; margin-right:{margin_right}em;"
        if 'style' in kwargs:
            style += kwargs.pop('style')

        tip: Tip = getattr(self.env.lang, button_name)
        tip_span = Html.tooltip(tip, tip.em)
        lvl_up   = self.env.level_up_from_current_page()
        img_link = get_ide_button_png_path(lvl_up,button_name)
        img      = Html.img(src=img_link, kls=HtmlClass.skip_light_box)

        button_html = Html.button(
            f'{ img }{ tip_span }{ extra_content }',
            btn_kind = btn_kind,
            kls = HtmlClass.tooltip,
            style = style,
            type = 'button',
                # https://developer.mozilla.org/en-US/docs/Web/HTML/Element/button#notes
            **kwargs,
        )
        return button_html
